// -*- C++ -*-
#pragma once

#include "defs.h"

#include "range.h"


namespace goopax {
namespace impl {
constexpr double PI = 3.14159265358979323846;


namespace powerplant {
template<size_t a>
struct highest_bit
{
    static const size_t mask = highest_bit<a / 2>::mask << 1;
};

template<>
struct highest_bit<1>
{
    static const size_t mask = 1;
};
template<>
struct highest_bit<0>
{
    static const size_t mask = 0;
};


template<class T, size_t pa>
struct A;

template<class T, size_t pa, size_t highbit>
struct A_help
{
    static constexpr T a(const T a)
    {
        return A<T, highbit>::a(a) * A<T, pa - highbit>::a(a);
    }
};
template<class T, size_t pa>
struct A_help<T, pa, pa>
{
    static constexpr T a(const T a)
    {
        return A<T, pa / 2>::a(a) * A<T, pa / 2>::a(a);
    }
};

// A::a(a) returns a to the power of the integer pa. (pa>0)
template<class T, size_t pa>
struct A : public A_help<T, pa, highest_bit<pa>::mask>
{
};

template<class T>
struct A<T, 0>
{
    static constexpr T a(const T)
    {
        return 1;
    }
};
template<class T>
struct A<T, 1>
{
    static constexpr T a(const T a)
    {
        return a;
    }
};


// B::b(a) returns the pb'th root of a. Add more specializations if appropriate.
template<class T, size_t pb>
struct B
{
    static inline constexpr T b(const T a)
    {
        // defaults to the slow std::pow function.
        return std::pow(a, static_cast<T>(1.0 / pb));
    }
};
template<class T>
struct B<T, 1>
{
    static inline constexpr T b(const T a)
    {
        return a;
    }
};
template<class T>
struct B<T, 2>
{
    static inline constexpr T b(const T a)
    {
        return sqrt(a);
    }
};
template<class T>
struct B<T, 3>
{
    static inline constexpr T b(const T a)
    {
        return cbrt(a);
    }
};
template<class T>
struct B<T, 4>
{
    static inline constexpr T b(const T a)
    {
        return sqrt(sqrt(a));
    }
};


template<int pa, unsigned int pb>
struct kuerzen;

template<int pa, unsigned int pb, bool div2, bool div3, bool div5>
struct kuerzen_help;

template<int pa, unsigned int pb, bool div3, bool div5>
struct kuerzen_help<pa, pb, true, div3, div5>
{
    // cancel by factor 2
    static const int A = kuerzen<pa / 2, pb / 2>::A;
    static const int B = kuerzen<pa / 2, pb / 2>::B;
};

template<int pa, unsigned int pb, bool div5>
struct kuerzen_help<pa, pb, false, true, div5>
{
    // cancel by factor 3
    static const int A = kuerzen<pa / 3, pb / 3>::A;
    static const int B = kuerzen<pa / 3, pb / 3>::B;
};

template<int pa, unsigned int pb>
struct kuerzen_help<pa, pb, false, false, true>
{
    // cancel by factor 5
    static const int A = kuerzen<pa / 5, pb / 5>::A;
    static const int B = kuerzen<pa / 5, pb / 5>::B;
};

template<int pa, unsigned int pb>
struct kuerzen_help<pa, pb, false, false, false>
{
    // cancellation not possible
    static const int A = pa;
    static const int B = pb;
};


template<int pa, unsigned int pb>
struct kuerzen
{
    static const int A = kuerzen_help<pa, pb, (pa % 2 == 0 && pb % 2 == 0), (pa % 3 == 0 && pb % 3 == 0), (pa % 5 == 0 && pb % 5 == 0)>::A;
    static const int B = kuerzen_help<pa, pb, (pa % 2 == 0 && pb % 2 == 0), (pa % 3 == 0 && pb % 3 == 0), (pa % 5 == 0 && pb % 5 == 0)>::B;
};


template<class T, int pa, unsigned int pb, bool pag0>
struct fracpower_help;

template<class T, int pa, unsigned int pb>
struct fracpower_help<T, pa, pb, true>
{
    // pa >= 0
    static inline constexpr T f(const T a)
    {
        return A<T, pa>::a(B<T, pb>::b(a));
    }
};

template<class T, int pa, unsigned int pb>
struct fracpower_help<T, pa, pb, false>
{
    // pa < 0
    static inline constexpr T f(const T a)
    {
        return T(1) / A<T, -pa>::a(B<T, pb>::b(a));
    }
};

template<class T, int pa, int pb>
inline constexpr T fracpower(const T a)
{
    // this returns the fractional power pa/pb of a.
    return fracpower_help<T, kuerzen<pa, pb>::A, kuerzen<pa, pb>::B, (pa >= 0)>::f(a);
}
}


template<int pa, int pb>
inline constexpr float pow(const float a)
{
    return powerplant::fracpower<float, pa, pb>(a);
}

template<int pa, int pb>
inline constexpr double pow(const double a)
{
    return powerplant::fracpower<double, pa, pb>(a);
}

template<int pa, int pb>
inline constexpr long double pow(const long double a)
{
    return powerplant::fracpower<long double, pa, pb>(a);
}

template<int pa, int pb, class T>
constexpr inline T pow(const T a)
{
    return powerplant::fracpower<T, pa, pb>(a);
}

template<int pa, int pb>
inline constexpr int pow(const int a)
{
    static_assert(pb == 1, "");

    return powerplant::fracpower<int, pa, pb>(a);
}


template<class T, int n>
struct powtype
{
    static T                           TDA;
    typedef decltype((pow<n, 1>(TDA))) type;
};

template<class T>
inline constexpr typename powtype<T, 2>::type pow2(const T& a)
{
    return pow<2, 1>(a);
}
template<class T>
inline constexpr typename powtype<T, 3>::type pow3(const T& a)
{
    return pow<3, 1>(a);
}
template<class T>
inline constexpr typename powtype<T, 4>::type pow4(const T& a)
{
    return pow<4, 1>(a);
}
template<class T>
inline constexpr typename powtype<T, 5>::type pow5(const T& a)
{
    return pow<5, 1>(a);
}
template<class T>
inline constexpr typename powtype<T, 6>::type pow6(const T& a)
{
    return pow<6, 1>(a);
}
template<class T>
inline constexpr typename powtype<T, 7>::type pow7(const T& a)
{
    return pow<7, 1>(a);
}
}
using impl::pow;
using impl::pow2;
using impl::pow3;
using impl::pow4;
using impl::pow5;
using impl::pow6;
using impl::pow7;
}
